# Copyright 2023 Agnostiq Inc.


from importlib import metadata

from .cloud_executor.cloud_executor import CloudExecutor
from .dispatch_management import dispatch, get_result, redispatch
from .qelectron_sdk.executors import CloudQCluster
from .service_account_interface.auth_config_manager import get_api_key, save_api_key
from .service_account_interface.client import get_client
from .shared.classes.settings import settings
from .swe_management.secrets_manager import delete_secret, list_secrets, store_secret
from .swe_management.swe_manager import create_env
from .volume.volume import volume

__version__ = metadata.version("covalent_cloud")



# Copyright 2023 Agnostiq Inc.


"""Unit tests for Dispatch management futures module."""

from unittest.mock import MagicMock

import covalent as ct

from covalent_cloud.dispatch_management.results_manager import (
    AssetScope,
    FutureResult,
    FutureVar,
    Result,
)


class TestFutureVar:

    MOCK_DISPATCH_ID = "mock-dispatch-id"
    MOCK_TASK_ID = 1
    MOCK_NAME = "mock-name"
    MOCK_VALUE = "mock-value"

    def test_init(self):
        """Test FutureVar init method."""

        mock_result_manager = MagicMock()

        future_var = FutureVar(
            mock_result_manager,
            AssetScope.RESULT,
            self.MOCK_NAME,
            self.MOCK_DISPATCH_ID,
            self.MOCK_TASK_ID,
        )
        assert future_var._dispatch_id == self.MOCK_DISPATCH_ID
        assert future_var._task_id == self.MOCK_TASK_ID
        assert future_var._name == self.MOCK_NAME
        assert future_var._value is None
        assert future_var._downloaded is False

    def test_load_result_scope(self):
        """Test FutureVar load for a result scope asset"""
        mock_result_manager = MagicMock()

        future_var = FutureVar(
            mock_result_manager,
            AssetScope.RESULT,
            self.MOCK_NAME,
            self.MOCK_DISPATCH_ID,
            None,
        )

        future_var.load()
        mock_result_manager.download_result_asset.assert_called()
        mock_result_manager.load_result_asset.assert_called()
        mock_result_manager.load_node_asset.assert_not_called()

    def test_load_lattice_scope(self):
        """Test FutureVar load for a lattice scope asset."""
        mock_result_manager = MagicMock()

        future_var = FutureVar(
            mock_result_manager,
            AssetScope.LATTICE,
            self.MOCK_NAME,
            self.MOCK_DISPATCH_ID,
            None,
        )

        future_var.load()
        mock_result_manager.download_lattice_asset.assert_called()
        mock_result_manager.load_lattice_asset.assert_called()

    def test_load_node_scope(self):
        """Test FutureVar load for a lattice scope asset."""
        mock_result_manager = MagicMock()

        future_var = FutureVar(
            mock_result_manager,
            AssetScope.NODE,
            self.MOCK_NAME,
            self.MOCK_DISPATCH_ID,
            self.MOCK_TASK_ID,
        )

        future_var.load()
        mock_result_manager.download_node_asset.assert_called()
        mock_result_manager.load_node_asset.assert_called()


class TestFutureResult:

    MOCK_DISPATCH_ID = "mock-dispatch-id"
    MOCK_PARENT_DISPATCH_ID = "mock-parent-dispatch-id"
    MOCK_DISPATCH_NAME = "mock-dispatch-name"
    MOCK_START_TIME = "mock-start-time"
    MOCK_END_TIME = "mock-end-time"
    MOCK_STATUS = "mock-status"
    MOCK_LATTICE = "mock-lattice"
    MOCK_RESULT = "mock-result"
    MOCK_INPUTS = "mock-inputs"
    MOCK_ERROR = "mock-error"

    MOCK_TASK_ID = 1

    def test_from_result_object(self):
        """Test FutureResult attribute properties."""

        dispatch_id = "test_from_result_object"

        @ct.electron
        def task(x):
            return x**2

        @ct.lattice
        def workflow(x):
            return task(x)

        workflow.build_graph(2)
        res = Result(workflow, dispatch_id)

        fr = FutureResult._from_result_object(res)
        assert fr.__dict__ == res.__dict__



# Copyright 2023 Agnostiq Inc.

import os
import platform
import time
from typing import Dict, List

import pennylane as qml
import requests
from covalent._serialize.common import AssetType, deserialize_asset, serialize_asset
from covalent._shared_files import logger

from covalent_cloud.qelectron_sdk.executors.base import CloudQCluster, CloudQExecutor
from covalent_cloud.shared.classes.exceptions import CovalentSDKError

from ....shared.classes.api import QElectronAPI
from ...schemas.schemas import CircuitCreateSchema, QExecutorSpecs
from .base_client import BaseQClient

app_log = logger.app_log

"""QClient for a remote QServer"""

qserver_url = os.environ.get("COVALENT_CLOUD_QSERVER_URL", "https://api.quantum.dev.covalent.xyz/")


class RemoteQClient(BaseQClient):

    """
    Client for a remote QServer.
    """

    RESULT_POLL_FREQ = 5  # seconds

    def __init__(self) -> None:

        self.api_client = QElectronAPI(qserver_url, url_prefix="/job/api/v1")

        # Since the cloud qserver only tracks circuits and not
        # batches, the onus of assembling execution results within a
        # batch falls on the qclient.
        self._circuit_ids_by_batch_id: Dict[str, List[str]] = {}

    @property
    def selector(self):
        return self.deserialize(self.qserver.selector)

    @selector.setter
    def selector(self, selector_func):
        self.qserver.selector = self.serialize(selector_func)

    @property
    def database(self):
        return self.deserialize(self.qserver.database)

    # This will POST to the /circuits.
    def register(self, qscripts, executors, qelectron_info, qnode_specs) -> List[Dict]:
        allowed_executors = []

        # Each executor is a Pydantic model.
        # The variable `executors` contains one `QExecutor` or `QCluster`.
        real_executor = executors[0]

        # Unpack QCluster into allowed_executors and selectors.
        if isinstance(real_executor, CloudQCluster):
            selector = real_executor.selector
            for executor in real_executor.executors:
                attrs = executor.model_dump()
                name = attrs.get("name")
                allowed_executors.append(QExecutorSpecs(name=name, attributes=attrs))
        elif isinstance(real_executor, CloudQExecutor):
            selector = "random"
            attrs = real_executor.model_dump()
            name = attrs.get("name")
            allowed_executors.append(QExecutorSpecs(name=name, attributes=attrs))
        else:
            raise CovalentSDKError("Invalid quantum executor")

        circuit_name = qelectron_info.name
        circuit_description = qelectron_info.description

        dispatch_id = os.getenv("COVALENT_DISPATCH_ID")
        node_id = int(os.getenv("COVALENT_CURRENT_NODE_ID"))

        request_body = CircuitCreateSchema(
            python_version=platform.python_version(),
            dispatch_id=dispatch_id,
            node_id=node_id,
            circuit_name=circuit_name,
            circuit_description=circuit_description or "",
            qnode_specs=qnode_specs,
            allowed_qexecutors=allowed_executors,
            num_circuits=len(qscripts),
            selector=selector,
        )
        res = self.api_client.post(
            "/circuits",
            request_options={
                "data": request_body.json(),
            },
        )
        return res.json()

    def upload_asset(self, asset: bytes, url: str):
        raise NotImplementedError

    def start(self, circuit_ids: List[str]):
        res = self.api_client.post(
            "/qdispatches", request_options={"json": {"circuit_ids": circuit_ids}}
        )
        return res

    def submit(
        self, qscripts: List[qml.tape.QuantumScript], executors, qelectron_info, qnode_specs
    ) -> str:
        """
        Submit a batch of circuits to the qserver.

            Circuit submission consists of three stages:

            1. Register the circuits by POSTIng to /circuits. Save the
            returned sequence of circuit_ids.

            2. Upload circuits (qscripts) and circuit strings to the URLs returned in
            the previous response.

            3. Start the circuits by POSTing to /qdispatch
        """
        circuit_schemas = self.register(qscripts, executors, qelectron_info, qnode_specs)
        # app_log.debug(f"circuit_schemas: {circuit_schemas}")
        circuit_ids = list(map(lambda circuit: circuit.get("circuit_id"), circuit_schemas))

        if len(circuit_ids) == 0:
            raise CovalentSDKError(
                "No circuits were returned by the qserver after submission, no valid batch id available.",
                "remote_qclient/no-circuits-created",
            )

        # Identify a batch of circuits by the first circuit ID in the batch.
        batch_id = circuit_schemas[0]["batch_id"]
        self._circuit_ids_by_batch_id[batch_id] = circuit_ids

        # Upload serialized circuits and diagrams to the presigned S3 URLs.
        for i, qscript in enumerate(qscripts):
            circuit_url = circuit_schemas[i]["assets"]["circuit"]["uri"]
            circuit_str_url = circuit_schemas[i]["assets"]["circuit_string"]["uri"]

            # Get serialized qscript and circuit diagram string.
            ser_circuit = self.serialize(qscript)
            circuit_str = qscript.draw().encode("utf-8")

            # Upload assets to S3 using presigned URLs.
            requests.put(circuit_url, data=ser_circuit, timeout=None).raise_for_status()
            requests.put(circuit_str_url, data=circuit_str, timeout=None).raise_for_status()

        self.start(circuit_ids)

        return batch_id

    def get_remaining_circuit_ids(self, batch_id: str) -> List[str]:
        """Returns circuit IDs for all pending results."""
        res = self.api_client.get(f"/circuits/batches/{batch_id}")
        circuits = res.json()
        remaining_circuit_ids = []
        for circuit in circuits:
            if circuit.get("status") not in ["COMPLETED", "FAILED"]:
                remaining_circuit_ids.append(circuit.get("circuit_id"))
        return remaining_circuit_ids

    def get_results(self, batch_id: str):
        # Get all circuits to query for results.
        remaining_circuit_ids = self._circuit_ids_by_batch_id[batch_id]
        while len(remaining_circuit_ids) > 0:
            remaining_circuit_ids = self.get_remaining_circuit_ids(batch_id)
            time.sleep(self.RESULT_POLL_FREQ)

        deserialized_results = []
        for circuit_id in self._circuit_ids_by_batch_id[batch_id]:

            # Get the circuit info.
            circuit_info = self.api_client.get(f"/circuits/{circuit_id}").json()

            # Check if the circuit completed successfully.
            if circuit_info["status"] != "COMPLETED":
                error_msg = circuit_info["error"]
                raise RuntimeError(f"Circuit {circuit_id} failed to execute:\n{error_msg}")

            # Download results.
            result_uri = circuit_info["assets"]["result"]["uri"]
            res = requests.get(result_uri, timeout=None)
            res.raise_for_status()

            deserialized_results.append(self.deserialize(res.content))

        return deserialized_results

    def serialize(self, obj):
        return serialize_asset(obj, AssetType.OBJECT)

    def deserialize(self, ser_obj):
        return deserialize_asset(ser_obj, AssetType.OBJECT)



# Copyright 2023 Agnostiq Inc.

import sys
import time
from typing import Callable, Optional

import requests

from .api import APIClient
from .exceptions import CovalentSDKError
from .settings import Settings, settings


def check_env_is_ready(
    env_name: str,
    settings: Settings = settings,
    wait: Optional[bool] = False,
    timeout: Optional[int] = 1800,
) -> Callable:
    """
    Checks if the environment is ready.

    Args:
        env_name: Name of the environment to check.
        settings: Settings object with the dispatcher URI.
        wait: If True, waits until the environment is ready before returning.
        timeout: Timeout in seconds for the environment to be ready.

    Returns:
        True if the environment is ready, False otherwise.
        Environment response if the environment is ready.

    """
    is_env_ready = False
    dispatcher_addr = settings.dispatcher_uri
    dispatcher_port = settings.dispatcher_port
    env_response = {}
    try:
        client = APIClient(host_uri=dispatcher_addr, settings=settings, port=dispatcher_port)
        endpoint = "/api/v2/envs/"
        request_options = {"params": {"name": env_name}}

        if wait:
            end_time = time.time() + timeout
            interval = 5
            print(f"Waiting for environment {env_name} to be ready...", end="", flush=True)
            while time.time() < end_time:
                is_env_ready, env_response = _get_env_status(
                    env_name, client, endpoint, request_options, wait
                )
                if is_env_ready:
                    break
                else:
                    time.sleep(interval)
                print(".", end="", flush=True)

            print()
            if not is_env_ready:
                raise CovalentSDKError(
                    f'Environment creation timed out, "{env_name}" is not ready.'
                )
        else:
            is_env_ready, env_response = _get_env_status(
                env_name, client, endpoint, request_options, wait
            )

    except requests.exceptions.HTTPError as e:
        print(e.response.text, file=sys.stderr)
        raise e

    return is_env_ready, env_response


def _get_env_status(
    env_name: str, client: APIClient, endpoint: str, request_options, wait: bool = False
):
    """
    Gets the status of the environment.

    Args:
        env_name: Name of the environment to check.
        client: APIClient object.
        endpoint: Endpoint to send the request to.
        request_options: Options to send with the request.
        wait: If True, waits until the environment is ready before returning.

    Returns:
        True if the environment is ready, False otherwise.
        Environment response if the environment is ready.

    """
    is_env_ready = False

    env_response = client.get(endpoint, request_options).json()

    if len(env_response["records"]) == 0:
        raise CovalentSDKError(f'Environment "{env_name}" does not exist.')

    if env_response["records"][0]["status"] == "READY":
        is_env_ready = True
    elif not wait:
        raise CovalentSDKError(f'Environment "{env_name}" is not ready.')

    return is_env_ready, env_response



# Copyright 2023 Agnostiq Inc.


"""Models for volumes"""
import re

from pydantic import BaseModel, field_validator

from covalent_cloud.shared.classes.exceptions import CovalentSDKError

# Regex pattern for a valid directory name
# ^(/)? - Optionally starts with '/'
# ([\w-]+)$ - Followed by word characters or hyphens, and must end here
VOLUME_NAME_PATTERN = r"^/?([\w-]+)$"


class BaseVolume(BaseModel):
    name: str

    @staticmethod
    def is_volume_name_valid(name):
        match = re.match(VOLUME_NAME_PATTERN, name)
        return match is not None

    @field_validator("name")
    def name_must_be_valid_directory(cls, v):
        match = re.match(VOLUME_NAME_PATTERN, v)
        if not match:
            raise CovalentSDKError(
                "Name must be a valid directory name without subdirectories", "volume/invalid-name"
            )
        return match.group(1)

    @property
    def path(self):
        return f"/{self.name}"

    def __str__(self):
        return self.path


class Volume(BaseVolume):
    """Volume model"""

    id: int
    user_id: str
    access_point_id: str
    name: str
    deployment_id: str



# Copyright 2023 Agnostiq Inc.

from typing import Optional

from covalent_cloud import get_client
from covalent_cloud.shared.classes.settings import Settings, settings


def store_secret(name: str, value: str, settings: Optional[Settings] = settings):
    """
    Stores a new secret in the Covalent Cloud. This function is used to securely save sensitive information like API keys, passwords, or other confidential data. Each secret is identified by a unique name.

    .. note:: There is a limit of 100 secrets per user account.

    Args:
        name (str): The name of the secret to be stored.
        value (str): The value of the secret.
        settings (Settings, optional): Configuration settings for connecting to the Covalent Cloud. Defaults to None, which uses the default settings.

    Returns:
        None: This function does not return any value but raises an error if the operation is unsuccessful.

    Examples:
        .. highlight:: python
        .. code-block:: python

            import covalent_cloud as cc

            # Store a new secret
            cc.store_secret(name="MY_SECRET_NAME", value="MY_SECRET_VALUE")
            # A secret with the name "MY_SECRET_NAME" and value "MY_SECRET_VALUE" is stored.
    """
    client = get_client(settings)
    body = {
        "name": name,
        "value": value,
    }
    r = client.post("/api/v2/secrets", request_options={"json": body})
    r.raise_for_status()


def list_secrets(settings: Optional[Settings] = settings):
    """
    Retrieves a list of all stored secrets in the Covalent Cloud. This function is useful for managing and verifying the secrets stored within the user's environment.

    Args:
        settings (Settings, optional): Configuration settings for connecting to the Covalent Cloud. Defaults to None, which uses the default settings.

    Returns:
        List[str]: A list containing the names of all stored secrets.

    Examples:
        .. highlight:: python
        .. code-block:: python

            import covalent_cloud as cc

            # Retrieve a list of all stored secrets
            secret_names = cc.list_secrets()
            # Returns a list of names of all secrets stored in the user's environment.
    """
    client = get_client(settings)
    r = client.get("/api/v2/secrets")
    r.raise_for_status()
    return r.json()["names"]


def delete_secret(name: str, settings: Optional[Settings] = settings):
    """
    Deletes a specific secret from the Covalent Cloud. Use this function to remove secrets that are no longer needed or should be updated.

    Args:
        name (str): The name of the secret to be deleted.
        settings (Settings, optional): Configuration settings for connecting to the Covalent Cloud. Defaults to None, which uses the default settings.

    Returns:
        None: This function does not return any value but raises an error if the operation is unsuccessful.

    Examples:
        .. highlight:: python
        .. code-block:: python

            import covalent_cloud as cc

            # Optionally, delete a secret if no longer needed
            cc.delete_secret(name="MY_SECRET_NAME")
            # Deletes the secret named "MY_SECRET_NAME".
    """
    client = get_client(settings)

    r = client.delete(f"/api/v2/secrets/{name}")
    r.raise_for_status()



# Copyright 2023 Agnostiq Inc.

from typing import Optional

from covalent_cloud.shared.classes.api import DispatcherAPI
from covalent_cloud.shared.classes.settings import Settings


def get_client(settings: Optional[Settings] = None) -> DispatcherAPI:
    """
    A factory / getter method for the Dispatcher API client.

    Args:
        None

    Returns:
        An instance of `DispatcherAPI` client.

    """

    return DispatcherAPI(settings=settings)



# Copyright 2023 Agnostiq Inc.

"""Module for Covalent Cloud dispatching and related functionalities."""


from copy import deepcopy
from functools import wraps
from typing import Callable, Union

import covalent as ct
from covalent._results_manager.result import Result
from covalent._workflow.lattice import Lattice

from covalent_cloud.shared.classes.exceptions import CovalentAPIKeyError
from covalent_cloud.shared.schemas.volume import Volume

from ..shared.classes.settings import Settings, settings
from . import results_manager as rm
from .helpers import fast_redispatch, register, start, validate_executors

API_KEY = "fake"  # pragma: allowlist secret


def dispatch(
    orig_lattice: Lattice,
    settings: Settings = settings,
    volume: Union[Volume, None] = None,
) -> Callable:
    """
    Dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.
    The dispatch function takes a Covalent workflow, also called a lattice, and sends it to the Covalent Cloud Server for execution. Once dispatched, the workflow runs on a the cloud and can be monitored using the dispatch ID in the application. The dispatch function returns a wrapper function that takes the inputs of the workflow as arguments. This wrapper function can be called to execute the workflow.

    Args:
        orig_lattice: The Covalent workflow to send to the cloud.
        volume: [optional] Volume instance
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        A wrapper function which takes the inputs of the workflow as arguments.

    Examples:

        .. highlight:: python
        .. code-block:: python

            # define a simple lattice
            import covalent_cloud as cc
            import covalent as ct

            # create volume [optional]
            volume = cc.volume("/myvolume")

            @ct.lattice
            def my_lattice(a: int, b: int) -> int:
                return a + b

            # dispatch the lattice and get the assigned dispatch ID
            dispatch_id = cc.dispatch(my_lattice, volume=volume)(2, 3)
    """

    dispatcher_addr = settings.dispatcher_uri

    @wraps(orig_lattice)
    def wrapper(*args, **kwargs) -> str:
        """
        Send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch id of the workflow.

        """

        try:
            lattice = deepcopy(orig_lattice)

            # Enabling task_packing for the build_graph call as it isn't
            # supported yet by OS covalent.
            old_task_packing = ct.get_config("sdk.task_packing")
            ct.set_config("sdk.task_packing", "true")
            lattice.build_graph(*args, **kwargs)
            ct.set_config("sdk.task_packing", old_task_packing)

            # Check that only CloudExecutors are specified.
            if settings.validate_executors and not validate_executors(lattice):
                raise ValueError("One or more electrons have invalid executors.")

            # TODO: Update register endpoint to take volume_id as a param and associate EFS to job_def
            dispatch_id = register(lattice, volume=volume, settings=settings)(*args, **kwargs)
            return start(dispatch_id, settings)

        except CovalentAPIKeyError as e:
            e.rich_print(level="error")

    return wrapper


def redispatch(
    dispatch_id: str,
    input_args: list,
    input_kwargs: dict,
    rebuild: bool = False,
    settings: Settings = settings,
):
    """
    Re-dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.

    Args:
        dispatch_id: The dispatch ID of the workflow to re-dispatch.
        input_args: The positional arguments of the workflow.
        input_kwargs: The keyword arguments of the workflow.
        rebuild: Whether to rebuild the workflow locally before re-dispatching. Defaults to False.
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        The dispatch ID of the re-dispatched workflow.

    """
    # if input_args is used, input_kwargs should be empty
    if input_args and input_kwargs:
        raise ValueError("input_args and input_kwargs cannot be used together")

    message = ""
    if rebuild:
        response = rebuild_redispatch(dispatch_id, input_args, input_kwargs, settings)
        redispatch_id = response["redispatch_id"]
        message = f"Rebuilt and re-dispatched workflow, original dispatch_id: {dispatch_id}, new dispatch_id: {redispatch_id}"
    else:
        response = fast_redispatch(dispatch_id, input_args, input_kwargs, settings)
        redispatch_id = response["redispatch_id"]
        message = f"Re-dispatched workflow, original dispatch_id: {dispatch_id}, new dispatch_id: {redispatch_id}\nPlease allow up to 60 seconds for the new dispatch to appear."

    print(message)

    return redispatch_id


def get_result(
    dispatch_id: str,
    wait: bool = False,
    settings: Settings = settings,
    *,
    status_only: bool = False,
) -> Result:
    """
    Gets the result of a Covalent workflow that has been dispatched to the cloud.

    This function retrieves the result of a dispatched Covalent workflow that has been executed on the cloud. The function takes the dispatch ID of the workflow and retrieves the results from the cloud. The result is returned as a `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Args:
        dispatch_id: The dispatch ID assigned to the workflow.
        wait: Controls how long the function waits for the server to return a result. If False, the function will not wait and will return the current status of the workflow. If True, the function will wait for the result to finish and keep retrying until the result is available.
        status_only: If True, only retrieves the status of the workflow and not the full result.
        settings: The Covalent settings to use for the request.

    Returns:
        A `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Examples:

        .. highlight:: python
        .. code-block:: python

            # define a simple lattice
            import covalent_cloud as cc
            import covalent as ct

            @ct.lattice
            def my_lattice(a: int, b: int) -> int:
                return a + b

            # dispatch the lattice and get the assigned dispatch ID
            dispatch_id = cc.dispatch(my_lattice)(2, 3)

            # get the result of the dispatched lattice
            result = cc.get_result(dispatch_id, wait=True)

            # print the final result of the lattice
            print(result.result)
    """

    return rm.get_result(
        dispatch_id=dispatch_id,
        wait=wait,
        settings=settings,
        status_only=status_only,
    )


def rebuild_redispatch(
    dispatch_id: str,
    input_args: list,
    input_kwargs: dict,
    settings: Settings = settings,
) -> Callable:
    """
    Re-dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.

    Args:
        dispatch_id: The dispatch ID of the workflow to re-dispatch.
        input_args: The positional arguments of the workflow.
        input_kwargs: The keyword arguments of the workflow.
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        The dispatch ID of the re-dispatched workflow.

    """

    try:
        # loading original dispatch
        original_dispatch = get_result(dispatch_id, wait=True)
        original_dispatch.result.load()

        # load lattice
        original_dispatch.lattice.workflow_function.load()
        original_dispatch.lattice.metadata["hooks"].load()

        # build lattice
        redispatch_lattice = ct.lattice(original_dispatch.lattice.workflow_function.value)
        redispatch_lattice.metadata = original_dispatch.lattice.metadata.copy()
        redispatch_lattice.metadata["hooks"] = original_dispatch.lattice.metadata[
            "hooks"
        ].value.copy()

        redispatch_id = dispatch(redispatch_lattice, settings=settings)(
            *input_args, **input_kwargs
        )

    except Exception as e:
        raise e

    return redispatch_id



# Copyright 2023 Agnostiq Inc.


from enum import Enum

from pydantic.dataclasses import dataclass


class GPU_TYPES(Enum):
    V100 = "v100"


@dataclass
class GPU:
    type: GPU_TYPES = GPU_TYPES.V100
    num: int = 0
